# Overview

This charm provides a space efficient mirror/cache for the Ubuntu software repository.  The service can be deployed as a single unit or scaled out to multiple units.  There are two components of the service, a Squid proxy for package files and an Apace server hosting static copies of the repository metadata.  The charm is responsible for keeping the metadata up to date and consistent.

# Packages

The package pool is served by a Squid caching proxy.  The package cache will download the package once and then keep it in the cache for up to 6 months.  Package files can be retained for this long because they will change once uploaded as the file relates to a specific version of the package.  Each unit of the charm peers with the other units in the charm to share cached files and avoiding downloads from the regular Ubuntu archive.

# Metadata mirror
The metadata, in the 'dists' directory, is a static mirror hosted by an Apache server.  The metadata files reference each other and contain checksums for each other.  Specifically, the 'Release' file in each distribution directory contains MD5, SHA1, and SHA256 sums for the rest of the files in the directory structure beneath it.  If these get out of sync with each other a client will report a 'Hash Sum Mismatch'.  To avoid this, these files are not served by a caching proxy, they are downloaded to a staging directory and check for consistency.  Once the inactive staging directory of metadata is ready, it replaces the current active metadata atomically on the unit.   When multiple units are deployed, this switch to a new metadata version is orchestrated by the charm to minimize the window where inconsistent metadata is served from the service.

# Validation
Validation is performed by reading a 'Release' file in a distribution subdirectory to gather the file sizes and SHA256 sums for each file in the sub-tree.  The tree is then walked to check each file size and SHA256 sum against the data from the 'Release' file.  If these do not match, the data is retrieved from the archive server again and checked.  Multiple attempts are made to provide a consistent copy of the updated metadata, but failure to do so will result in no update to the active metadata for that two hour window.

# Single units
When deployed as a single unit, this charm will install apache and squid on a unit and immediately mirror and validate the metadata.  Once ready, the charm will begin serving metadata and providing access to the package proxy.  Every two hours the charm will download the changed metadata (using rsync) to an inactive copy of the metadata.  Once validation is completed it becomes the active copy of the metadata.

# Scaling
When additional units are added to the service they will download and validate metadata from the source archive.  Once complete they will join the cluster for the service and pull the latest metadata from the leader (the active unit with the lowest unit ID number).  Every two hours the leader will update metadata and push it to all known units in the service.  Then it will store the new metadata version number to the Juju peer unit relationship for the charm.  Each unit will run their 'cluster-relationship-changed' hook, see the new metadata version and switch to serving the new metadata.  If the peer unit does not have a copy of the metadata, it will shut down apache, pull the latest metadata from the leader, and then restart apache to begin serving the newest metadata.  A unit may need to pull metadata if the leader was unable to successfully push metadata due to transient network issues, or if the peer joined the cluster after the leader began its operation; in this case the leader would have been unaware of the new peer.

# Leadership
The leader of the cluster is simply the lowest numbered active unit.  It will have a cron job which runs every 2 hours to fetch and validate metadata and then push the update to the other peers.  Thus the leader is responsible for minimizing load on the main archive, ensuring up-to-date, consistent metadata, and notifying the rest of the cluster that  new metadata is available.

